unit UInspGrid;
interface
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls;

const
MaxColCount = 3;
WinOffSet = 2;

type
  TButtonStyle = (btNone, btArrow, btElipsis);

  TCustomInspEdit = class(TCustomEdit)
  private
    { Private declarations }
    fOldText      : TCaption;
    fButtonWidth : integer;
    fButtonStyle : TButtonStyle;
    fPressed : boolean;
    fTracking : boolean;
    fBtnRect : TRect;
    fOnButtonClick : TNotifyEvent;
    fCanvas : TCanvas;
    procedure StopTracking;
    procedure SetButtonStyle(Value: TButtonStyle);

    procedure TrackButton(X,Y: Integer);
    procedure ArrowDown(ARect : Trect; P :boolean);
    procedure Elipsis(ARect : TRect; p :Boolean);
    procedure WMPaint(var Message: TWMPaint); message wm_Paint;
    procedure WMSize(var Msg :TWMSize);message WM_Size;
  protected
    { Protected declarations }
    procedure BoundsChanged;
    procedure CreateParams(var Params: TCreateParams); override;

    procedure CNkeydown(var Message: TWMkeydown); message CN_KEYDOWN;
    procedure CMEnter( Var Message : TCMEnter );   message CM_ENTER;
    procedure CMExit( Var Message : TCMExit );   message CM_EXIT;

    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure PaintWindow(DC: HDC); override;
    procedure EditButtonClick;

  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property ButtonStyle: TButtonStyle read fButtonStyle write SetButtonStyle default btNone;
    property CharCase;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font ;
    property HideSelection;
    property MaxLength;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Text;
    property Visible;
    property OnButtonClick : TNotifyEvent read fOnButtonClick write fOnButtonClick;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  published
    { Published declarations }
  end;

TCustomInspLBox = class(TCustomListBox)
 public
  { Public declarations }
 constructor Create(AOwner: TComponent); override;
end;

TInspLBox = class(TCustomInspLBox)
end;



  TInspEdit = class(TCustomInspEdit)
 published
    { Published declarations }
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property ButtonStyle;
    property CharCase;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font ;
    property HideSelection;
    property MaxLength;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Text;
    property Visible;
    property OnButtonClick;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

//==== TCustomInspGrid    =================
TDrawState = (ds_Flat,ds_Lowered);

TCellType = (ctViewer, ctShowFocus, ctShowEdit, ctCallEditor,
                                          ctCallList, ctEdit );

TSetCellTypeEvent  = procedure(Sender: TObject; ARow :word;
            var CType :TCellType)of object;

TCallListBoxEvent = procedure(Sender: TObject; ARow :word;
                             var List : string) of object;

TCustomInspGrid = class(TScrollBox)
private
    { Private declarations }
fColWidth  : array[0..2] of word;
fCol              : word;
fColCount         : byte; { maximum value 3 }
fDefaultColWidth  : word;
fDefaultRowHeight : word;
fRow              : word;
fRowCount         : word;
fCellType         : TCellType;
fEdit             : TCustomInspEdit;
fListBox          : TCustomInspLBox;
fOnSetCellType    : TSetCellTypeEvent;
fEditRow          : word;
fOnBtnClick       : TNotifyEvent;
fOnCallListbox    : TCallListBoxEvent;
fVGrRect          : TRect;
fVisibleRowCount  : word;
fTopRow           : word;
 fDrLine  : boolean;
 xOld     : word;    {stores position of cursor for MoseMove }
 xCol     : word;    {stores Column where drawing was started }
 fCanvas    : TCanvas;
 fItems     : TStringList; {stores Cells content}
function GetGridRectangle : TRect;
function GetLeftCoordinate(Acol : byte):word;
procedure fSetLastColWidth;
procedure UpdateScrollRange;
procedure SetColCount(aValue: byte);
procedure SetDefaultColWidth(aValue: word);
function GetColWidth(index : byte):word;
procedure SetColWidth(index : byte; value : word);
procedure SetRowCount(aValue: word);
procedure SetDefaultRowHeight(aValue: word);
procedure SetRow(AValue : word);
procedure SetCellType(aValue : TCellType);
procedure DrawBevel(ARect : TRect;AState : TDrawState);
procedure DrawCellText(ACol, ARow : word;  ARect :TRect);
procedure DrawCell(ACol, ARow : word; AState : TDrawState);virtual;
procedure DrawRow(ARow : word);
procedure WMPaint(var Message : TWMPaint); message WM_Paint;
procedure InsertEditor(aRow : Word);
procedure DeActivateEditor( aRow : Word);
//=========================================================================
procedure InitItemsList;
function GetCells(ACol,ARow : word) : string;
procedure SetCells(ACol,ARow : word; const AValue : string);
procedure DeleteRowString(ARow,Num : integer);
procedure InsertRowString(ARow,Num : integer);
procedure DeleteColString(ACol,Num : byte);
procedure InsertColString(ACol,Num : integer);
function GetObjects(ACol,ARow : word) : TObject;
procedure SetObjects(ACol,ARow : word; const AValue : TObject);
//=========================================================================
procedure MouseToCell(X, Y: Integer; var ACol, ARow: word);
procedure MouseDown(Button:TMouseButton; Shift:TShiftState; X,Y:Integer); override;
procedure MouseUp(Button:TMouseButton; Shift:TShiftState; X,Y:Integer); override;
procedure MouseMove(Shift: TShiftState; X, Y: Integer);override;
procedure WMkeydown(var Message: TWMkeydown);message WM_KEYDOWN;
procedure WMCommand(var M : TWMCommand);message WM_COMMAND;
procedure fEditBtnClick;
//=========================================================================
procedure SetListBoxBounds;
procedure InitFListBox;
procedure FreeListBox;
//=========================================================================
procedure WMGetDlgCode (var Message : TMessage); message WM_GETDLGCODE;
procedure WMSize(var Msg : TWMSize); message WM_Size;
  protected
    { Protected declarations }
procedure WndProc(var message : TMessage); override;
procedure PaintWindow(DC: HDC); override;
property Canvas : TCanvas read fCanvas;
  public
    { Public declarations }
constructor Create(AOwner: TComponent); override;
procedure Loaded; override;
destructor Destroy; override;
procedure InsertRows(ARow,Num : integer); virtual;
procedure DeleteRows(ARow,Num : integer); virtual;
function CellRect(aCol,ARow : word):TRect;
property ColWidth[Index : byte]:word read GetColWidth write SetColWidth;
 property VisibleRowCount : word read fVisibleRowCount;
 property Cells[ACol,ARow : word] : String read GetCells write SetCells;
 property Objects[ACol, ARow: word]: TObject read GetObjects write SetObjects;
 property Row : Word read fRow write SetRow default 0;
 property ColCount :byte read fColCount write SetColCount default 2;
 property DefaultColWidth  : word read fDefaultColWidth
                       write SetDefaultColWidth default 64;
 property DefaultRowHeight : word read fDefaultRowHeight
                        write setDefaultRowHeight  default 22;
 property RowCount : word read fRowCount write SetRowCount default 12;
 property CellType : TCellType read fCellType write SetCellType default ctShowFocus;
 property OnSetCellType : TSetCellTypeEvent read fOnSetCellType write fOnSetCellType;
 property OnBtnClick : TNotifyEvent read fOnBtnClick write fOnBtnClick;
 property OnCallListBox : TCallListBoxEvent read fOnCallListbox write fOnCallListbox;
  published
    { Published declarations }
 property TabStop default True;
  end;

TInspGrid = class(TCustomInspGrid)
published
 property ColCount;
 property DefaultRowHeight;
 property RowCount;
 property Align;
 property DragCursor;
 property DragMode;
 property Enabled;
 property Font;
 property ParentFont;
 property ParentShowHint;
 property TabStop;
 property Visible;
 property OnBtnClick;
 property OnCallListBox;
 property OnDragDrop;
 property OnDragOver;
 property OnEndDrag;
 property OnEnter;
 property OnExit;
 property OnSetCellType;
 property OnMouseDown;
 property OnMouseMove;
 property OnMouseUp;
 property OnStartDrag;
end;

procedure Register;

implementation
{$R UInspGrid.Res}

procedure Register;
begin
  RegisterComponents('Samples', [TInspGrid]);
end;
{=========================================================}
constructor TCustomInspEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fButtonWidth := GetSystemMetrics(SM_CXVSCROLL);
  fButtonStyle := btNone;
  Visible := True;
fCanvas := TControlCanvas.Create;
TControlCanvas(fCanvas).Control := Self;
end; // Create

destructor TCustomInspEdit.Destroy;
begin
fCanvas.Free;
  inherited Destroy;
end; // Destroy

procedure TCustomInspEdit.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
if fButtonStyle <> btNone then
  with Params do  begin
    Style := Style or ES_MULTILINE;
  end;
end;  // CreateParams

procedure TCustomInspEdit.BoundsChanged;
var
  R: TRect;
begin
 SetRect(R, 0, 0, ClientWidth - 2, ClientHeight + 1); // +1 is workaround for windows paint bug
  if (fButtonStyle <> btNone)then begin
SetRect(fBtnRect, ClientWidth - fButtonWidth, 0, ClientWidth, ClientHeight);
          Dec(R.Right, fButtonWidth);
                                               end;
  SendMessage(Handle, EM_SETRECT, 0, LongInt(@R));
  Repaint;
end; // BoundsChanged

procedure TCustomInspEdit.SetButtonStyle(Value: TButtonStyle);
var oldValue: TButtonStyle;
begin
if Value = fButtonStyle then Exit;
 oldValue:= fButtonStyle;
  fButtonStyle := Value;
if HandleAllocated  then begin
   if ((Ord(OldValue) = 0) and (Ord(fButtonStyle) > 0))
     or ((Ord(OldValue) > 0) and (Ord(fButtonStyle) = 0))
  then  RecreateWnd;
         BoundsChanged; end;
end; // SetLinkStyle


procedure TCustomInspEdit.EditButtonClick;
begin
  if Assigned(fOnButtonClick) then fOnButtonClick(Self);
    if HasParent then PostMessage(Parent.Handle,WM_COMMAND,BN_CLICKED,0);
end; // EditButtonClick

procedure TCustomInspEdit.WMPaint(var Message: TWMPaint);
begin
  PaintHandler(Message)
end; // WMPaint

procedure TCustomInspEdit.CNkeydown(var Message: TWMkeydown);
var
  Msg        : TMsg;
  fKeyAction : integer;
begin
   Case Message.CharCode of
VK_RETURN :
 if  (fButtonStyle <> btNone)
   and (ssCtrl In KeyDataToShiftState(Message.KeyData)) then
  begin
    EditButtonClick;
    PeekMessage(msg, Handle, WM_CHAR, WM_CHAR, PM_REMOVE);
  end  else fKeyAction := VK_DOWN;
VK_DOWN   : fKeyAction := VK_DOWN;
VK_UP     : fKeyAction := VK_UP;
VK_ESCAPE : begin Text := fOldText; fKeyAction := VK_ESCAPE; end;
 else inherited;
  end;
  if HasParent then PostMessage(Parent.Handle,WM_KEYDOWN,fKeyAction,0 )
end;


procedure TCustomInspEdit.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  WasPressed: Boolean;
begin
 WasPressed := fPressed;
  StopTracking;
  if (Button = mbLeft) and (fButtonStyle <> btNone) and WasPressed then
    EditButtonClick;
  inherited MouseUp(Button, Shift, X, Y);
end;

procedure TCustomInspEdit.ArrowDown(ARect : Trect; P :boolean);
var m1,m3,v1 : word;
    OldColor : TColor;
begin
OldColor :=  fCanvas.Brush.Color;
 if p then fCanvas.Brush.Color := clGray else
             fCanvas.Brush.Color := clBlack;
with ARect do begin
m1  := Left + ((Right - Left) shr 1);
m3 :=  ((Right - Left) shr 2);
v1 :=  (Top + (Bottom - Top) shr 1) +  Ord(p);
 fCanvas.Polygon([Point( m1 - m3, v1 - m3), Point(m1 + m3, v1 - m3),
    Point(m1, v1 + m3)]);
end;
  fCanvas.Brush.Color := OldColor;
end;

procedure TCustomInspEdit.Elipsis(ARect : TRect; p :Boolean);
var w,fl : integer;
begin
with ARect do begin
fl := ((Right - Left) shr 1) - 1 + Ord(p);
w :=  (Right - Left) shr 3;
 if w = 0  then w := 1;
PatBlt(fCanvas.Handle,Left + fl,Top + fl,w,w,BLACKNESS);
PatBlt(fCanvas.Handle,Left + fl-W*2,Top + fl,w,w,BLACKNESS);
PatBlt(fCanvas.Handle,Left + fl+W*2,Top + fl,w,w,BLACKNESS);
 end;
end;

procedure TCustomInspEdit.WMSize(var Msg :TWMSize);
begin
   inherited;
BoundsChanged;
end;


procedure TCustomInspEdit.PaintWindow(DC: HDC);
var
  Flags: Integer;
  W: Integer;
begin
if (fButtonStyle <> btNone) then
 begin
   Flags := 0;
if FPressed then Flags := BF_FLAT;
    DrawEdge(DC, fBtnRect, EDGE_RAISED, BF_RECT or BF_MIDDLE or Flags);
if fButtonStyle = btElipsis then Elipsis(fBtnRect,fPressed)
 else   ArrowDown(fBtnRect,fPressed);
with fBtnRect do ExcludeClipRect(DC, Left, Top, Right, Bottom);
end;
Text := Trim(Text);
  inherited PaintWindow(DC);
end; // PaintWindow

procedure TCustomInspEdit.StopTracking;
begin
if FTracking then
  begin
    TrackButton(-1, -1);
    FTracking := False;
    MouseCapture := False;
  end;
end; // StopTracking;


procedure TCustomInspEdit.TrackButton(X,Y: Integer);
var
  NewState: Boolean;
begin
 NewState := PtInRect(fBtnRect, Point(X, Y));
  if fPressed <> NewState then
  begin
    fPressed := NewState;
    InvalidateRect(Handle, @fBtnRect, False);
  end;
end; // TrackButton


procedure TCustomInspEdit.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
 if (Button = mbLeft) and (fButtonStyle <> btNone) and Focused
  and PtInRect(fBtnRect, Point(X,Y)) then
    begin
    MouseCapture := True;
    FTracking := True;
    TrackButton(X, Y);
    end;
  inherited MouseDown(Button, Shift, X, Y);
end; // MouseDown

procedure TCustomInspEdit.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
if (fButtonStyle <> btNone)
  and PtInRect(fBtnRect,Point(X,Y)) then  Screen.Cursor:= crArrow else
       Screen.Cursor:= crDefault;
  if fTracking then
    TrackButton(X, Y);
  inherited MouseMove(Shift, X, Y);
end; // MouseMove


procedure TCustomInspEdit.CMExit( Var Message : TCMExit );
begin
  if (fButtonStyle <> btNone) then BoundsChanged;
 Modified := False;
if HasParent then Parent.SetFocus;
  inherited;
 end;

procedure TCustomInspEdit.CMEnter( Var Message : TCMEnter );
begin
 inherited;
 fOldText := Text;    {stores Text }
  if (fButtonStyle <> btNone)
    then BoundsChanged;
end;

{========= TCustomInspLBox =================}
constructor TCustomInspLBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
BorderStyle := bsNone;
Visible := False;
end;


{============= TCustomInspGrid =================}

function TCustomInspGrid.GetGridRectangle : TRect;
 begin
  fVisibleRowCount := ClientHeight div fDefaultRowHeight;
  if fVisibleRowCount > fRowCount then
                   fVisibleRowCount := fRowCount;
Height := fVisibleRowCount * fDefaultRowHeight + 4;
  Result := ClientRect;
  UpdateScrollRange;
if (fEdit <> nil) then InSertEditor(fEditRow);
end;

function TCustomInspGrid.GetLeftCoordinate(Acol : byte):word;
var i : byte;
begin
 Result := 0;
if aCol > 0 then
 for i := 0 to ACol - 1 do
  Result := Result + fColWidth[i];
end;

procedure TCustomInspGrid.fSetLastColWidth;
var  wdth  : word;
begin
wdth :=  ClientRect.Right - ClientRect.Left;
if GetLeftCoordinate(fColCount - 1) <  wdth then
 fColWidth[ ColCount - 1 ] := wdth - GetLeftCoordinate(fColCount - 1) - 2
    else  fColWidth[ fColCount - 1 ] := 0;
end;

procedure TCustomInspGrid.UpdateScrollRange;
var h : word;
begin
with VertScrollBar do begin
 Increment := fDefaultRowHeight;
  Position := 0;
   fTopRow := 0;
  h := ClientRect.Bottom - Clientrect.Top;
   Range := h+(fRowCount - fVisibleRowCount)*fDefaultRowHeight;
                      end;
end;

procedure TCustomInspGrid.WndProc(var message : TMessage);
var aTop : word;
begin
case  Message.msg of
 WM_VScroll :
    begin
    aTop := GetScrollPos(Handle,SB_VERT) div fDefaultRowHeight;
    if aTop <> fTopRow then begin
     fTopRow := aTop;
     Invalidate;
                            end;
      end;
  end;
  inherited WndProc(message);
end;

procedure TCustomInspGrid.SetColCount(aValue : byte);
var i,delta : byte;
begin
if aValue <= MaxColCount then
  if (fColCount <> aValue) then
 begin
  if fColCount > aValue then
     DeleteColString(AValue, fColCount - aValue);
  if fColCount < aValue then
     InsertColString(fColCount, aValue - fColCount);
 fColCount := aValue;
   for i := 0 to ColCount -2 do
    fColWidth[i] := fDefaultColWidth;
  fSetLastColWidth;
if (fEdit <> nil) and (fRow = fEditRow)
 then  InSertEditor(fRow);
 Invalidate;
 end;
end;

procedure TCustomInspGrid.SetDefaultColWidth(aValue: word);
var i : byte;
 begin
if fDefaultColWidth <> aValue then begin
 fDefaultColWidth := aValue;
   for i := 0 to ColCount -2 do
    fColWidth[i] := fDefaultColWidth;
       fSetLastColWidth;
  Invalidate;
 end;
end;

function TCustomInspGrid.GetColWidth(index : byte):word;
begin
  Result := fColWidth[index]
end;

procedure TCustomInspGrid.SetColWidth(index : byte; value : word);
begin
if fColWidth[index] <> value   then
 begin
  fColWidth[index] := value;
    fSetLastColWidth;
      Repaint;
  end;
 end;



procedure TCustomInspGrid.SetDefaultRowHeight(aValue: word);
begin
  if (fDefaultRowHeight <> aValue) then
 begin
 fDefaultRowHeight := aValue;
  fVGrRect := GetGridRectangle;
   Invalidate;
 end;
end;

procedure TCustomInspGrid.SetRowCount(aValue: word);
begin
if fRowCount <> aValue then
 begin
if fRowCount < aValue then
  InsertRowString(fRowCount,aValue-fRowCount);
if fRowCount > aValue then
 DeleteRowString(aValue,fRowCount-aValue);
    fVGrRect := GetGridRectangle;
Repaint
 end;
end;



procedure TCustomInspGrid.SetRow(AValue : word);
begin
if fRow <> aValue then
begin
   fRow := aValue;
if Assigned(FOnSetCellType) then
          FOnSetCellType(Self, fRow, FCellType);
 Invalidate;
 end;
end;

procedure TCustomInspGrid.SetCellType(aValue : TCellType);
begin
 if fCellType <> AValue then
 begin
            fCellType := AValue;
if Ord(fCelltype) > 1 then begin
  if HandleAllocated and (fEdit = nil) then begin
       fEdit := TCustomInspEdit.Create(TCustomInspGrid(Owner));
        InsertControl(fEdit);
                                             end;
with  fEdit do begin
      ReadOnly := True;
      Visible := False;
      Color := clWindow;
      BorderStyle := bsSingle;
Case fCellType of
ctShowEdit   : ButtonStyle := btNone;
ctCallEditor : ButtonStyle := btElipsis;
ctCallList   : ButtonStyle := btArrow;
ctEdit       : begin ButtonStyle := btNone; ReadOnly := False; end;
              end;
                 end; {with fEdit}
                              end{Ord(fCelltype) > 0}

end;
end;

procedure TCustomInspGrid.DrawBevel(ARect : TRect; AState : TDrawState);
begin
 with ARect,Canvas do
 begin
 Pen.Style:= psInsideFrame;
if AState = ds_Flat then begin
 Pen.Color := clBtnShadow;
    PolyLine([ Point(Left, Bottom),Point(Right-1, Bottom),Point(Right-1, Top)]);
   Pen.Color := clBtnHighlight;
   if Left > 0 then
      PolyLine([ Point(Left, Bottom),Point(left, Top)]);
                         end;
if AState = ds_Lowered then begin
  Pen.Color := clBlack;
  if Left = 0  then
   PolyLine([Point(Left,Bottom), Point(Left, Top),Point(Right, Top)])
       else  PolyLine([Point(Left, Top),Point(Right, Top)]);
   Pen.Color := clBtnHighlight;
      PolyLine([ Point(Left, Bottom-1),Point(Right, Bottom-1),Point(Right,Top)]);
                           end;
 end;
end;

procedure TCustomInspGrid.DrawCellText(ACol, ARow : Word;  ARect :TRect);
var
 clText     : array[0..79] of Char;
 R          : TRect;
 FHeight    : Integer;
  begin
with fCanvas do begin
Brush.Color := Color;
 if (ARow = fRow) then  Font.Color  := clWindowText
      else   Font.Color  := clGrayText;
  R :=ARect;
   InflateRect(R,-1,-1);
    FillRect(R);
     FHeight := TextHeight('W');
  with R do begin
   Top := ((Bottom + Top) - FHeight) shr 1;
   Bottom := Top + FHeight;
   Inc(Left,2);
            end;
   StrPlCopy(clText,Cells[ACol, ARow],Length(Cells[ACol, ARow]));
    DrawText(Handle, clText, StrLen(clText), R, (DT_EXPANDTABS or
       DT_VCENTER or DT_Left));
  end; {with fCanvas}
end;

procedure  TCustomInspGrid.DrawCell(ACol, ARow : word; AState : TDrawState);
var ARect : TRect;
   Skeep : boolean;
begin
 if ARow >= fTopRow then
  begin
   Skeep := False;
    aRect := CellRect(ACol,ARow);
if (ACol = 1) and (fEdit <> nil) then begin
if (fEditRow < fRowCount) then
 if (aRow = fRow) and (fRow <>  fEditRow) then  begin
         DeActivateEditor(fEditRow);
                                      end;
  if (ARow = fRow) and (fRow <>  fEditRow)
     and (Ord(fCelltype) > 1) then begin
       InSertEditor(aRow);
              Skeep := (ARow = fEditRow) and (ACol = 1);
                                   end;
end;
 if not Skeep then begin
      DrawBevel(ARect,AState);
 DrawCellText(ACol, ARow, ARect);
                    end;
  end{ ARow >= fTopRow}
end;

procedure  TCustomInspGrid.DrawRow(ARow : word);
var     i : byte;
   AState : TDrawState;
begin
if (ARow >= fTopRow) and (ARow < RowCount) then begin
 if (fCellType <> ctViewer) and
   (ARow = fRow) then  AState :=  ds_Lowered else  AState :=  ds_Flat;
 for i:= 0 to fColCount - 1 do DrawCell( i, ARow, AState );
  end;
end;

procedure TCustomInspGrid.WMPaint(var Message : TWMPaint);
begin
  PaintHandler(Message);
end;

procedure TCustomInspGrid.PaintWindow(DC: HDC);
 var i : byte;
begin
for i := fTopRow to fTopRow + fVisibleRowCount do
 begin
     DrawRow(i);
 end;
  inherited PaintWindow(DC);
end; // PaintWindow


procedure  TCustomInspGrid.InSertEditor( aRow : Word);
var r : Trect;
begin
  R := CellRect(1,ARow);
    InflateRect(R,0,-1);
with R do fEdit.SetBounds(Left,Top,Right - Left,Bottom - Top);
if HandleAllocated then
with fEdit do begin
 Visible := True;
  Text := Cells[1,ARow];
  SetFocus;
                end;
   fEditRow := aRow;
end;

procedure  TCustomInspGrid.DeActivateEditor( aRow : Word);
var aRect : TRect;
begin
if fEdit.HandleAllocated then begin
  Cells[1,ARow] := fEdit.Text;
  fEdit.Visible := False;
  fEditRow := fRowCount + 1; {out from RowCount}
   DrawCell(1, ARow,ds_Flat);
  end;
end;

procedure TCustomInspGrid.InitItemsList;
var i : longint;
begin
for i := 0 to fColCount * fRowCount do
   fItems.Add('');
end;

function TCustomInspGrid.GetCells(ACol,ARow:word) : string;
begin
 if (ACol <= fColCount-1) and (ARow <= fRowCount-1) then
   Result := fItems.Strings[fRowCount * ACol + ARow];
end;

procedure TCustomInspGrid.SetCells(ACol,ARow:word; const AValue : string);
begin
  if (ACol <= fColCount-1) and (ARow <= fRowCount-1) then
   begin
      fItems.Strings[fRowCount * ACol + ARow] := AValue;
     if ARow >= fTopRow then DrawRow(ARow)
   end;
end;

procedure TCustomInspGrid.DeleteRowString(ARow,Num : integer);
var i,j     : byte;
 OldCount : longint;
begin
 if ARow + num <= fRowCount then begin
 OldCount :=   fItems.Count; {stores initial fItems.Count}
 for i := 1 to fColCount do
   for j := 1 to num do
     fItems.Delete(OldCount - fRowCount * i + ARow);
            Dec(fRowCount,num);
                  end;
end;

procedure TCustomInspGrid.InsertRowString(ARow,Num : integer);
var i,j     : byte;
 OldCount : longint;
begin
 OldCount :=  fItems.Count;
 for i := 1 to fColCount do
   for j := 1 to num do
     fItems.Insert(OldCount - fRowCount * i + ARow,'');
            Inc(fRowCount,num);
end;

procedure TCustomInspGrid.DeleteColString(ACol,Num : byte);
var i : byte;
 fromItem,ToItem     : word;
begin
 fromItem := ACol * fRowCount;
  toItem := fromItem + fRowCount * Num;
     if toItem <= fItems.Count then
 for i := toItem -1 downto fromItem  do
        fItems.Delete(i)
 end;

procedure TCustomInspGrid.InsertColString(ACol,Num : integer);
var i : byte;
 fromItem,ToItem     : word;
begin
  fromItem := ACol * fRowCount;
  toItem := fromItem + fRowCount * Num;
 for i := fromItem to toItem-1  do
        fItems.Insert(i,'')
end;

function TCustomInspGrid.GetObjects(ACol,ARow : word) : TObject;
begin
   Result := fItems.Objects[ACol*ARow];
end;

procedure TCustomInspGrid.SetObjects(ACol,ARow : word; const AValue : TObject);
begin
if AValue <> nil
then fItems.InsertObject(ACol*ARow, fItems.Strings[ACol*ARow], AValue);
end;

procedure TCustomInspGrid.MouseToCell(X, Y: Integer; var ACol, ARow: word);
var i : byte;
    w : word;
begin
ARow := Y div fDefaultRowHeight + fTopRow;
  i := 0;
   w := 0;
  while  W < X  do begin
   W := W + fColWidth[i];
            Inc(i)
                   end;
ACol := i-1;
end;

procedure TCustomInspGrid.MouseDown(Button:TMouseButton; Shift:TShiftState; X,Y:Integer);
begin
if FListBox <> nil then FreeListBox;
if (Button = mbLeft)  then
 if (Screen.Cursor = crHSplit) then
  begin
   XOld := X;
     Canvas.MoveTo(xOld,0);
     Canvas.LineTo(xOld,ClientHeight);
  end else
   begin
 if Ord(fCelltype) < 1 then SetFocus;
{ new fRow  } MouseToCell(X, Y,fCol,fRow);
if Assigned(FOnSetCellType) then  FOnSetCellType(Self, fRow, FCellType);
      Invalidate;
   inherited;
   end;
end;

procedure TCustomInspGrid.MouseUp(Button:TMouseButton; Shift:TShiftState; X,Y:Integer);
var oldW_1 : word;
begin
if fDrLine then begin
 fColWidth[xCol] := X - CellRect(xCol,1).Left;
 fDrLine := False;
 Screen.Cursor:= crDefault;
  fSetLastColWidth;
if (fEdit <> nil) and (fRow = fEditRow)
 then  InSertEditor(fRow);
     Invalidate;
  end else
  begin
    inherited;
  end;
 end;

procedure TCustomInspGrid.MouseMove(Shift: TShiftState; X, Y: Integer);
var mRow  : word;
   cRect  : TRect;
begin
if PtInRect(ClientRect, Point(X, Y))and not fDRLine then
begin
 MouseToCell(X,Y,xCol,mRow);
      CRect := CellRect(xCol,mRow);
if ((X + 5 >= CRect.Right) or
 ((X - 5 <= CRect.Left)) and (xCol < ColCount-1)
  and (xCol > 0)) and (X + 5 < ClientRect.Right) then
 begin
  Screen.Cursor:= crHSplit;
  if (Shift = [ssLeft]) then
        begin
        fDrLine := True;
   if (X-5 <= CRect.Left) then Dec(xCol);
         end;
       end   else Screen.Cursor:= crDefault;
end {ClientRect}
 else Screen.Cursor:= crDefault;
if fDrLine then   begin
 Screen.Cursor:= crIBeam;
Canvas.Pen.Mode := pmNotXor;
  Canvas.Pen.Color := clBtnShadow;
   Canvas.MoveTo(xOld,0);
  Canvas.LineTo(xOld,ClientHeight);
 Canvas.MoveTo(x,0);
  Canvas.LineTo(x,ClientHeight);
                   end;
Xold := x;
Canvas.Pen.Mode := pmCopy;
   inherited;
end;

procedure TCustomInspGrid.WMkeydown(var Message: TWMkeydown);
var
  key     : word;
  oldfRow : word;
begin
Key := Message.CharCode;
oldfRow := fRow;
case Key of
 VK_DOWN  : if (fRow < RowCount-1) then
    begin
      Inc(fRow);
   if fRow >= fTopRow + fVisibleRowCount then  begin
     Inc(fTopRow);  {and do scroll}
      SetScrollPos(Handle,SB_VERT,fTopRow*fDefaultRowHeight,True) ;
                                             end;
      end;
VK_UP : if fRow > 0 then
begin
 Dec(fRow);
   if FRow < fTopRow then begin
    Dec(fTopRow);
        SetScrollPos(Handle,SB_VERT,fTopRow*fDefaultRowHeight,True) ;
   end;
 end;
  else
    inherited;
    end;  {Case of}
if fRow <> oldfRow then begin
if Assigned(FOnSetCellType) then  FOnSetCellType(Self, fRow, FCellType);
if Assigned(OnClick) then OnClick(Self);
      Invalidate;
                       end;
end;

procedure TCustomInspGrid.WMCommand(var M : TWMCommand);
begin
   if M.NotifyCode = BN_CLICKED then fEditBtnClick;

end;

procedure TCustomInspGrid.fEditBtnClick;
begin
if (fCellType = ctCallList) then
  if (FListBox = nil) then InitFListBox
    else FreeListBox
else
    if Assigned(fOnBtnClick) then fOnBtnClick(Self);
end;

procedure TCustomInspGrid.WMGetDlgCode (var Message : TMessage);
begin
   Message.result :=  DLGC_WANTARROWS;
end;

procedure TCustomInspGrid.InsertRows(ARow,Num : integer);
begin
  InsertRowString(ARow,Num);
    fVGrRect := GetGridRectangle;
      Invalidate;
end;

procedure TCustomInspGrid.DeleteRows(ARow,Num : integer);
begin
  DeleteRowString(ARow,Num);
    fVGrRect := GetGridRectangle;
      Invalidate;
end;

function TCustomInspGrid.CellRect(aCol,ARow : word):TRect;
var     ALeft, ATop : word;
begin
try
 ALeft := GetLeftCoordinate(Acol);
  ATop := (ARow - FTopRow) * fDefaultRowHeight;
Result :=  Bounds(ALeft, ATop, fColWidth[ACol], fDefaultRowHeight);
except on EAccessviolation  do
 end;
end;

procedure TCustomInspGrid.SetListBoxBounds;
var R     : Trect; {item rectangle of fListBox}
    Y,      {Top of fListBox}
    X,      {Left Coordinate of fListbox}
    dY,     {distance from fListBox.Top to InspGrid.Bottom}
    Hgt,    {height of fListBox }
    MWidth, {maximum Length(Strings[i])  }
    W
        : word;
    i   : byte;
begin
if fListBox.HandleAllocated  then
with fListBox do begin
 Y := FEdit.Top + FEdit.Height+ 2;
   MWidth := ColWidth[1];   {initial fListBox.Width}
     Hgt  :=  18;             {initial fListBox.height }
 if (Items.Count>0) then   begin
  Perform(LB_GETITEMRECT,0,LongInt(@R));
     Hgt := (R.Bottom - R.Top) * fListBox.Items.Count;
       dY :=  fVGrRect.Bottom - fVGrRect.Top - Y;
         if (dY < Hgt) then Hgt := dY;
{find fListBox width}
   for i := 0 to Items.Count-1 do  begin
         w := fCanvas.TextWidth(Items[i]);
           if  w > MWidth  then MWidth := w;
                                   end;
 if MWidth > Parent.Width then MWidth := Parent.Width;
    end;
      X := fEdit.Left + Fedit.Width - MWidth;
  SetBounds(x, y, MWidth,Hgt);
  end; {HandleAllocated}
end;

procedure TCustomInspGrid.InitFListBox;
var lst : string;
begin
 Lst := '';
if Assigned(fOnCallListbox) then fOnCallListbox(Self,fRow,Lst);
 fListBox := TCustomInspLBox.Create(TCustomInspGrid(Owner));
        InsertControl(fListBox);
with fListBox do begin
 if Lst <> '' then Items.CommaText := Lst;
   SetListBoxBounds;
    Visible := TRue;
      SetFocus;
               end;
end;


procedure TCustomInspGrid.FreeListBox;
begin
with FListBox do
 if ItemIndex >= 0  then begin
  FEdit.Text := Items[ItemIndex];
      Cells[1,fRow] := fEdit.Text;
                         end;
  FListBox.Free;
FListBox := nil;
end;

procedure TCustomInspGrid.WMSize(var Msg :TWMSize);
begin
 inherited;
  fVGrRect := GetGridRectangle;
           fSetLastColWidth;
end;


constructor TCustomInspGrid.Create(AOwner: TComponent);
var i : byte;
begin
  inherited Create(AOwner);
Color:= clBtnFace;
fColCount := 2;
fDefaultColWidth := 64;
 for i := 0 to fColCount - 1 do fColWidth[i] := fDefaultColWidth;
fRowCount := 12;
fDefaultRowHeight := 22;
fTopRow := 0;
fRow := 0;
fDrLine  := false;
 fCellType := ctViewer;
 fEditRow := fRowCount + 1;

fCanvas := TControlCanvas.Create;
TControlCanvas(fCanvas).Control := Self;
 fItems := TStringList.Create;
         InitItemsList;
end; // Create


procedure TCustomInspGrid.Loaded;
begin
  inherited Loaded;
fVGrRect := GetGridRectangle;
     fSetLastColWidth;
end;

destructor TCustomInspGrid.Destroy;
begin
 FItems.Clear;
 FItems.Free;
 FCanvas.Free;
 inherited Destroy;
end;

end.


